@@ -1,7 +1,7 @@
 use hir::{db::HirDatabase, HasSource, HasVisibility, PathResolution};
 use ide_db::base_db::FileId;
 use syntax::{
-    ast::{self, VisibilityOwner},
+    ast::{self, HasVisibility as _},
     AstNode, TextRange, TextSize,
 };
 
@@ -46,8 +46,7 @@ fn add_vis_to_referenced_module_def(acc: &mut Assists, ctx: &AssistContext) -> O
     let current_module = ctx.sema.scope(path.syntax()).module()?;
     let target_module = def.module(ctx.db())?;
 
-    let vis = target_module.visibility_of(ctx.db(), &def)?;
-    if vis.is_visible_from(ctx.db(), current_module.into()) {
+    if def.visibility(ctx.db()).is_visible_from(ctx.db(), current_module.into()) {
         return None;
     };
 
@@ -148,7 +147,7 @@ fn target_data_for_def(
     ) -> Option<(TextSize, Option<ast::Visibility>, TextRange, FileId)>
     where
         S: HasSource<Ast = Ast>,
-        Ast: AstNode + ast::VisibilityOwner,
+        Ast: AstNode + ast::HasVisibility,
     {
         let source = x.source(db)?;
         let in_file_syntax = source.syntax();
